import unittest
import os
import tempfile

from .ToolTest import ToolTest


# If the REPAIR_FAILURES environment variable is set, any tests failing due to wrong
# output will be corrected. Be sure to do a "git diff" to validate that you're
# getting changes you expect.

REPAIR_FAILURES = os.getenv("REPAIR_FAILURES", 0)


def get_test_cases():
    TESTS_PATH = os.path.join(os.path.dirname(__file__), "test_cases")
    paths = []
    for dirpath, dirnames, filenames in os.walk(TESTS_PATH):
        for fn in filenames:
            if fn.endswith(".txt") and fn[0] != '.':
                paths.append(os.path.join(dirpath, fn))
    paths.sort()
    test_cases = []
    for p in paths:
        with open(p) as f:
            # allow "#" comments at the beginning of the file
            comments = []
            while 1:
                cmd = f.readline()
                if cmd[0] != '#':
                    break
                comments.append(cmd)
            expected_output = f.read()
            test_name = os.path.relpath(
                p, TESTS_PATH).replace(".", "_").replace("/", "_")
            test_cases.append((test_name, cmd, expected_output, comments, p))
    return test_cases


class CmdlineTest(ToolTest):
    pass


def make_f(cmd, expected_output, comments, path):

    def f(self):
        CACHE_DIR = tempfile.mkdtemp()
        old_environ = dict(os.environ)
        new_environ = dict(PYCOIN_CACHE_DIR=CACHE_DIR)
        for k in "PATH PYCOIN_BTC_PROVIDERS".split():
            new_environ[k] = os.environ.get(k, "")
        os.environ = new_environ
        os.chdir(CACHE_DIR)
        for c in cmd.split(";"):
            actual_output = self.launch_tool(c)
        if actual_output != expected_output:
            print(cmd)
            print(actual_output)
            print(expected_output)
            if REPAIR_FAILURES:
                f = open(path, "w")
                f.write(''.join(comments))
                f.write(cmd)
                f.write(actual_output)
                f.close()
        os.environ = old_environ
        self.assertEqual(expected_output, actual_output)
    return f


def inject():
    for idx, (name, i, o, comments, path) in enumerate(get_test_cases()):
        name_of_f = "test_%s" % name
        setattr(CmdlineTest, name_of_f, make_f(i, o, comments, path))
        print("adding %s" % name_of_f)


inject()


def main():
    unittest.main()


if __name__ == "__main__":
    main()



import binascii
import unittest

from pycoin.encoding.hexbytes import b2h, h2b_rev
from pycoin.symbols.btc import network

Tx = network.tx

TX_E1A18B843FC420734DEEB68FF6DF041A2585E1A0D7DBF3B82AAB98291A6D9952_HEX = (
    "0100000001a8f57056b016d7d243fc0fc2a73f9146e7e4c7766ec6033b5ac4cb89c64e"
    "19d0000000008a4730440220251acb534ba1b8a269260ad3fa80e075cd150d3ffba76a"
    "d20cd2e8178dee98b702202284f9c7eae3adfcf0857a901cd34f0ea338d5744caab88a"
    "fad5797be643f7b7014104af8385da9dc85aa153f16341a4015bc95e7ff57876b9bde4"
    "0bd8450a5723a05c1c89ff2d85230d2e62c0c7690b8272cf85868a0a0fc02f99a5b793"
    "f22d5c7092ffffffff02bb5b0700000000001976a9145b78716d137e386ae2befc4296"
    "d938372559f37888acdd3c71000000000017a914c6572ee1c85a1b9ce1921753871bda"
    "0b5ce889ac8700000000")


address_for_script = network.address.for_script


class TxTest(unittest.TestCase):

    def test_tx_api(self):
        tx = Tx.from_hex(TX_E1A18B843FC420734DEEB68FF6DF041A2585E1A0D7DBF3B82AAB98291A6D9952_HEX)
        # this transaction is a pay-to-hash transaction
        self.assertEqual(tx.id(), "e1a18b843fc420734deeb68ff6df041a2585e1a0d7dbf3b82aab98291a6d9952")
        address = address_for_script(tx.txs_out[0].puzzle_script())
        self.assertEqual(address, "19LemzJ3XPdUxp113uynqCAivDbXZBdBy3")
        address = address_for_script(tx.txs_out[1].puzzle_script())
        self.assertEqual(address, "3KmkA7hvqG2wKkWUGz1BySioUywvcmdPLR")

    def test_blanked_hash(self):
        tx = Tx.from_hex(TX_E1A18B843FC420734DEEB68FF6DF041A2585E1A0D7DBF3B82AAB98291A6D9952_HEX)
        self.assertEqual(tx.id(), "e1a18b843fc420734deeb68ff6df041a2585e1a0d7dbf3b82aab98291a6d9952")
        self.assertEqual(
            b2h(tx.blanked_hash()), "909579526c4c2c441687c7478d3f96249724d2ff071d2272b44500d6cf70d5d6")
        tx.txs_in[0].script = b"foo"
        self.assertEqual(
            b2h(tx.blanked_hash()), "909579526c4c2c441687c7478d3f96249724d2ff071d2272b44500d6cf70d5d6")
        tx.txs_out[0].coin_value += 1
        self.assertEqual(
            b2h(tx.blanked_hash()), "10d4e87f7bf35f2949e7693e7a4a84189aad8631f0b2b0999e88f7261066cbe5")
        tx.txs_in[0].script = b"bar"
        self.assertEqual(
            b2h(tx.blanked_hash()), "10d4e87f7bf35f2949e7693e7a4a84189aad8631f0b2b0999e88f7261066cbe5")
        tx.txs_in[0].script = b""
        self.assertEqual(b2h(tx.hash()), "10d4e87f7bf35f2949e7693e7a4a84189aad8631f0b2b0999e88f7261066cbe5")
        tx.txs_in[0].script = b"foo"
        self.assertEqual(b2h(tx.hash()), "c91910058722f1c0f52fc5c734939053c9b87882a9c72b609f21632e0bd13751")

    def test_issue_39(self):
        """
        See https://github.com/richardkiss/pycoin/issues/39 and
        https://github.com/richardkiss/pycoin/pull/40

        There was a problem validating the following transactions:

        315ac7d4c26d69668129cc352851d9389b4a6868f1509c6c8b66bead11e2619f
        dbf38261224ebff0c455c405e2435cfc69adb6b8a42d7b10674d9a4eb0464dca
        de744408e4198c0a39310c8106d1830206e8d8a5392bcf715c9b5ec97d784edd

        This codes tests this.
        """
        TX_B64_LIST = [
            # some encoded transactions (the three listed above and the three
            # that they depend upon)
            (
                "AQAAAALcOOk1m9faO1g4YgThhtlAhoX0J/XlE2ZttzWqimshaQAAAABqRzBE"
                "AiBdj+6zEkeORo0LUU5j4ROVjXIU+lcqzYcHmn8MwCb8XAIgD6duoFvyQ69t"
                "D5F38kHK9gbQH8/V5i1r77yiTlaeXCcDIQIQChqcosGJMtZXfFjyJVgBhNDg"
                "gibUGVmHSslj48Gy/v/////cOOk1m9faO1g4YgThhtlAhoX0J/XlE2ZttzWq"
                "imshaQEAAABrSDBFAiAIft44cp5tNeT1FVBQGOZZIiAxJztzZpIPOT7jqxe8"
                "HgIhAMpDFkt1fRptEjXxMgDUtfdt2P2k7J/ChUay31sSEejfAyECdZg5E+YA"
                "k7dn6FWXypOX+y9Bjlf5mNavu8U2EWCFscv/////AUCJlQAAAAAAGXapFPzJ"
                "s204z1XX1bTuTd22ssF2EvSMiKwAAAAA"
            ),
            (
                "AQAAAAEtUf3HWib/PGE4Ag4am7QPH6tuOc6W/q4yGMmuA14AqwEAAABrSDBF"
                "AiEA5PGlIZB+UPxE0zEy7pjJcVpk350sKGDj4EdMUhq4U34CIDCvjTUGpTUu"
                "KwVkRazYVaQtNycOlKYpp7KLIYcOxtdhASEDgIxJPwYZkNK+AB5A8EiuiHAy"
                "C3SJXOLZZS88HHPNbyz/////AvCHSwAAAAAAGXapFPzJs204z1XX1bTuTd22"
                "ssF2EvSMiKzwh0sAAAAAABl2qRQzzvYXSdEboq3wkaXgRWeBd/46bYisAAAA"
                "AA=="
            ),
            (
                "AQAAAAJa+fLO2OCiRk98qhSvobvRyPsY3qrl0QEZa1jcIn70rgAAAABqRzBE"
                "AiANKFITbLHEu93eBOx29YHRsyockZFIyF+8D9BWXTWK8wIgNvKqF87Ind6w"
                "A3aigYv3KMRHmSgLnyBExWkad7Dc2WwDIQIQChqcosGJMtZXfFjyJVgBhNDg"
                "gibUGVmHSslj48Gy/v////9a+fLO2OCiRk98qhSvobvRyPsY3qrl0QEZa1jc"
                "In70rgEAAABrSDBFAiEA9APIYMTjztPlIyyzWCXnk3It+vCsLwGWGpN4K0kG"
                "qWMCIGLdifJz5mvPrW8FqLDNJrp7Bma+/Qw9pF2feVcX2lBKAyECdZg5E+YA"
                "k7dn6FWXypOX+y9Bjlf5mNavu8U2EWCFscv/////AaAClAAAAAAAGXapFOUK"
                "XY2jOZUbBAutBFPXxAz9dNPciKwAAAAA"
            ),
            (
                "AQAAAAGfYeIRrb5mi2ycUPFoaEqbONlRKDXMKYFmaW3C1MdaMQAAAABsSTBG"
                "AiEAhIisrGQ/6Sa7DAJtv+pa9nMiHuBTLNAkxlyzDjYvGEQCIQCFH27K+zjJ"
                "ItZHnrCORpOhrBnHvPnUX8mqXy1pGB/4ngEhAhAKGpyiwYky1ld8WPIlWAGE"
                "0OCCJtQZWYdKyWPjwbL+/////wKgxEoAAAAAABl2qRT8ybNtOM9V19W07k3d"
                "trLBdhL0jIisoMRKAAAAAAAZdqkUM872F0nRG6Kt8JGl4EVngXf+Om2IrAAA"
                "AAA="
            ),
            (
                "AQAAAALCBkSoNGHOnUgtcCy8I87ODdMmW1WL56GNNOIWvaccAAAAAABrSDBF"
                "AiAxKffbGKLs4sDhPFwLZvQlHX+Q20uxr0hFzQqtnSQZQAIhAImY0R1z7HrT"
                "Tt4hR0R/3n3eS8LXk14G94/O8Pc7LDlmAyECE2UQ39BTBuo0mCvz395yuOSd"
                "QyqYBb9kUtOZTnkvnRn/////yRF9O6xy+bn8PWf3KNM1uywKHCYWOL0bgEe1"
                "Zd1jGaIAAAAAakcwRAIgRQ7h/BpT6uurhfpEmEE/Xx5OAZdUohj+Euzr3Zg8"
                "mbkCIDxIakZ02TMLAtt5OHKyy0VQw7uywxjyis6540zeNZdJAyED78tvrsro"
                "6386Jta3YJd/I64guTuYS8oof9K4PDGZeHD/////AeD9HAAAAAAAGXapFB0x"
                "6lo758/yr1vtc3EOtvXV9n1wiKwAAAAA"
            ),
            (
                "AQAAAAKerCh2TFeXmFaXU1qdQUucoCL5WRFVNZdvNt1FZgp5XQAAAACMSTBG"
                "AiEAvLz97Qz/zSlKSDrllLRwj73G2B7RfaiR1ZspOG5Ae3kCIQD5ATZgiNvH"
                "X8Tn8Ib8RohgW0HGbPRi00XUcvxCTmybGgFBBCsXId9LDBz91gENMCmVXxRE"
                "ZI+E6QOSkToVTtny7tiOJhmHy/jci4KzQmucvUBotsK5r4CiwjhjOkAAXRD6"
                "SWD/////6864dM1/4fxjvltUc0HJ1da9agsSw4LV3KYhGR7FJ+MBAAAAi0gw"
                "RQIhAJIopjUy7dPOHa+LGTvgM4jfZ8pA522/Jx3+uFC4Lz5IAiBzLNoxejaa"
                "dw1CXwOUuzI4rMl0xsuYC5XQaxZNT2TFzwFBBBPpriULEjb9VdVoC8v3E4is"
                "RMmfQByPCJYadSwK/ZZg9TTFGyDXUwW+dQ9tScDzhMWfdLK9DyV4iAbnYh/S"
                "2cr/////A0BCDwAAAAAAGXapFFzGycfh13x6rrUPhNJNj2ViE7xbiKwACT0A"
                "AAAAABl2qRQhQVEH8cwnc3//rGPcfvakBANJxIistBcsAAAAAAAZdqkUMQV+"
                "QpfDgBAsCQ+ixaUK5Kgl0kOIrAAAAAA="
            ),
            (
                "AQAAAAO1CFlm1mEB3fjCtilQEH+6TbR3UzdJyqafj3mab9Mc6gAAAACKRzBE"
                "AiA8rWZ4BB8YYJp3xtx8jAZdrfQ6B0zjYRdgTS7I5LZF7gIgabCjn9iu9L3n"
                "YvKrdXFJJygtbg6V8iMTLrPh8ghdGvwBQQQrFyHfSwwc/dYBDTAplV8URGSP"
                "hOkDkpE6FU7Z8u7YjiYZh8v43IuCs0JrnL1AaLbCua+AosI4YzpAAF0Q+klg"
                "/////8IGRKg0Yc6dSC1wLLwjzs4N0yZbVYvnoY004ha9pxwAAQAAAItIMEUC"
                "IDNZYWLuCV0nJL6CCGgUfQfNoh0oAACd2lMZn+zJdJCDAiEAqZafa18G1K1x"
                "/6yOvj8h1uAGSM8UjSJJ6479li5sos4BQQTswrqYR5m+x0vFTzgGrrM2k+Gx"
                "gX+hDBAvN8Kq9RRuWdqC4jVNGhGdFD63Ev1TQYXMqvp6b9ztbAZ3ED8i6sFo"
                "/////0Vf19DzvUs2DvFwlVW9viTF+YlXCNYNMD6yUXK9I9RBAgAAAItIMEUC"
                "IQCKbaQY2eH1fsXZFksstrP4B+uxPBwGRe2Wxl7rW5sYGwIgVvVEPdnJNvVj"
                "rh0XZdhqnOAA0Sw39Upqkejrm+yXWnwBQQQ1hDJBuzoTc1ZJ8zyVQjEfRcjW"
                "o8rq3lE+3x3rYZ3Q/9xBEBtsnkFAzps/N8n6C5cK2QAmRGxeGFmbYaGFT5RP"
                "/////wNAQg8AAAAAABl2qRSU70Qwi2d2bI+nKnCP19XGsbSnWoisVEkwAAAA"
                "AAAZdqkUgroT7ai54LzKPXVnWJsPoV6lJ0yIrHjrFQAAAAAAGXapFEFyZV9I"
                "izJXnWmTivO2n9OKDWCdiKwAAAAA"
            ),
            (
                "AQAAAAHBHumhtHyFj2ma501AFchO/RrrfkY1sYTKsJiYe6i5pAEAAADaAEcw"
                "RAIgJQsMj5xe4yyGSQOseNBu7zuQNbdwYRpmu4tyOeVrDhoCIHTRJ5lHr5OH"
                "JsmDYl4nTEMhT2TeEN8tMNtrt/rFLMaHAUgwRQIhAObKZ2o5NubR2aoXKP7q"
                "oNMI3sv4u33Hnxcu1NBCilhoAiAH5OaEGAC5snVQDIWgXXVWICosFmTHHjXg"
                "y5fNwAO5gAFHUiECzr9qtYCUjRRrfMdx2OZGl0NJ09exHz4DKH0Jl6R307kh"
                "A3umUUhbeiyyIhketkpVkm5iu6v+m17SqUiKrVR7IEKCUq7/////An1KDwAA"
                "AAAAGXapFNxnIa33YyARGtMFwzhMdn1LmeGViKxllyYPAAAAABepFNsSg3N8"
                "2T68HrEpjWRKeEbFWm2WhwAAAAA="
            ),
            (
                "AQAAAAHZI2Rm7Gvz7UMEKi20P7AIT5AOxlhwW29S0uFz9EPz1QEAAADaAEgw"
                "RQIhAIX1NZuYzrKUHFAxUNYI6yWMUuzCEapuZOUY6TdCspWaAiAchzgPP6if"
                "WNh0cmVkyW1UpygM/eVa1XrtHepCMhvArAFHMEQCIGLJtKtbyJEaH6iQS+hK"
                "xUlGrWwmqdJScz/JfSZ1Qln6AiBNRC+gswEEMjTNR5uVetqCGJkNL2m6fDfk"
                "DyICU/otoQFHUiECzr9qtYCUjRRrfMdx2OZGl0NJ09exHz4DKH0Jl6R307kh"
                "A3umUUhbeiyyIhketkpVkm5iu6v+m17SqUiKrVR7IEKCUq7/////Aux5CAAA"
                "AAAAGXapFDIKbLrYWAn/2ZTB7ToisbIaZ5DoiKzL5TUPAAAAABepFNsSg3N8"
                "2T68HrEpjWRKeEbFWm2WhwAAAAA="
            ),
            (  # 837dea37ddc8b1e3ce646f1a656e79bbd8cc7f558ac56a169626d649ebe2a3ba
                "AQAAAAGsp/O0VlTCMOCIalf7mIwwRO9ej385cm0wXGHV6BiQPAAAAAD9XQEAS"
                "DBFAiABh6+Sjp0VXEsaycHJEYFTI5q6dndPd118H5w+EG/zPAIhAIgisPZY7e"
                "wiJ00LauneEOvy2gaxu9qrpOUOsHjznj14AUcwRAIgeV8PT1lBp3rgMuy54zd"
                "TeI1+tcsMeNgFV11rAKHZv+0CID4fStkzLRQWrgHicDjpRbydtZxzJyijg6bx"
                "7S+5naekAUzJUkEEkbuiUQkSpb032h+1sWcwEOQ9LG2BLFFOkb+p8usSnhwYM"
                "ynbVb2GjiCarC+8Assz2Y/nS/I/DCNdYSax2DNPhkEEhlxAKTpoDLnAIOex4Q"
                "bYwZFtPO+ZqkMaVtJT5pJW2sCe8SKxqYaBiny2JFMvBiwdH4ciCEhhxcMpHM/"
                "+9OxodEEEjSRV0kA+CHCPwfVWAC8bbNg/mS0IUJf5l0qwiiiDjweJb7qwjzlJ"
                "XhX6b61u2/sedU41+hx4RMQfMioYY9RiE1Ou/////wFAQg8AAAAAABl2qRSuV"
                "rTbE1VNMhxALbOWEYeu0bvtW4isAAAAAA=="
            ),
            (  # 3c9018e8d5615c306d72397f8f5eef44308c98fb576a88e030c25456b4f3a7ac
               # input of
               # 837dea37ddc8b1e3ce646f1a656e79bbd8cc7f558ac56a169626d649ebe2a3ba
                "AQAAAAGJYyhI+ZcikVcnxcddqNstvxlDQqBCmCj2b/iPqyr31gAAAACLSDBFA"
                "iEAq7yKc/4gVEgL2j8ygdotDFHihBORq9TAn0+QiiA0wY0CIFvJ5NaOr7kY8+"
                "lmIzhkekQZwN4aZQq4mD8dIW4qMdjjAUEEb1XXre/2ARx+rClP5UDFeDC+gOk"
                "1XIOGnJJgpLi/R2ema6y9cLgE3GPVvusUGAKSrX87CDNysdAtejfdl/9cnv//"
                "//8BQEIPAAAAAAAXqRT4FbA22bu85enyoAq9G/PckelVEIcAAAAA"
            )
        ]

        TX_LIST = [Tx.from_hex(b2h(binascii.a2b_base64(b64.encode("utf8")))) for b64 in TX_B64_LIST]
        TX_DB = dict((tx.hash(), tx) for tx in TX_LIST)
        for h in ["315ac7d4c26d69668129cc352851d9389b4a6868f1509c6c8b66bead11e2619f",
                  "dbf38261224ebff0c455c405e2435cfc69adb6b8a42d7b10674d9a4eb0464dca",
                  "de744408e4198c0a39310c8106d1830206e8d8a5392bcf715c9b5ec97d784edd",
                  "485716e53b422aca0fe5b1ded21360695ce5f49255d80e10db56458ed6962ff3",
                  "837dea37ddc8b1e3ce646f1a656e79bbd8cc7f558ac56a169626d649ebe2a3ba"]:
            tx = TX_DB.get(h2b_rev(h))
            self.assertNotEqual(tx, None)
            tx.unspents_from_db(TX_DB)
            for idx, tx_in in enumerate(tx.txs_in):
                self.assertTrue(tx.is_solution_ok(tx_in_idx=idx))


def tx_to_b64(tx_hex):
    # use this to dump raw transactions in the data above
    tx = Tx.from_hex(tx_hex)
    d = tx.as_bin()
    for idx in range(0, len(d), 45):
        print('"%s"' % binascii.b2a_base64(d[idx:idx+45]).decode("utf8")[:-1])


if __name__ == "__main__":
    unittest.main()



from pycoin.networks.bitcoinish import create_bitcoinish_network


network = create_bitcoinish_network(
    network_name="Monacoin", symbol="TMONA", subnet_name="testnet4",
    wif_prefix_hex="ef", sec_prefix="TMONASEC:", address_prefix_hex="6f", pay_to_script_prefix_hex="75",
    bip32_prv_prefix_hex="04358394", bip32_pub_prefix_hex="043587cf", bech32_hrp="tmona",
    magic_header_hex="fdd2c8f1", default_port=19403,
    dns_bootstrap=["testnet-dnsseed.monacoin.org"])



import io

from .hash import sha256
from pycoin.coins.bitcoin.SegwitChecker import ZERO32
from pycoin.coins.bitcoin.SolutionChecker import BitcoinSolutionChecker
from pycoin.encoding.bytes32 import from_bytes_32

from pycoin.satoshi.satoshi_struct import stream_struct

from pycoin.satoshi.flags import (
    SIGHASH_NONE, SIGHASH_SINGLE, SIGHASH_ANYONECANPAY,
)


class GroestlcoinSolutionChecker(BitcoinSolutionChecker):
    def _hash_prevouts(self, hash_type):
        if hash_type & SIGHASH_ANYONECANPAY:
            return ZERO32
        f = io.BytesIO()
        for tx_in in self.tx.txs_in:
            f.write(tx_in.previous_hash)
            stream_struct("L", f, tx_in.previous_index)
        return sha256(f.getvalue())

    def _hash_sequence(self, hash_type):
        if (
                (hash_type & SIGHASH_ANYONECANPAY) or
                ((hash_type & 0x1f) == SIGHASH_SINGLE) or
                ((hash_type & 0x1f) == SIGHASH_NONE)
        ):
            return ZERO32

        f = io.BytesIO()
        for tx_in in self.tx.txs_in:
            stream_struct("L", f, tx_in.sequence)
        return sha256(f.getvalue())

    def _hash_outputs(self, hash_type, tx_in_idx):
        txs_out = self.tx.txs_out
        if hash_type & 0x1f == SIGHASH_SINGLE:
            if tx_in_idx >= len(txs_out):
                return ZERO32
            txs_out = txs_out[tx_in_idx:tx_in_idx+1]
        elif hash_type & 0x1f == SIGHASH_NONE:
            return ZERO32
        f = io.BytesIO()
        for tx_out in txs_out:
            stream_struct("QS", f, tx_out.coin_value, tx_out.script)
        return sha256(f.getvalue())

    def _signature_for_hash_type_segwit(self, script, tx_in_idx, hash_type):
        return from_bytes_32(sha256(self._segwit_signature_preimage(script, tx_in_idx, hash_type)))



import io
import re

from binascii import b2a_base64, a2b_base64

from pycoin.intbytes import byte2int, int2byte
from pycoin.satoshi.satoshi_string import stream_satoshi_string

from ..encoding.bytes32 import to_bytes_32, from_bytes_32
from ..encoding.exceptions import EncodingError
from ..encoding.hash import double_sha256
from ..encoding.sec import public_pair_to_hash160_sec


class MessageSigner(object):

    # According to brainwallet, this is "inputs.io" format, but it seems practical
    # and is deployed in the wild. Core bitcoin doesn't offer a message wrapper like this.
    signature_template = ('-----BEGIN {net_name} SIGNED MESSAGE-----\n{msg}\n-----BEGIN '
                          'SIGNATURE-----\n{addr}\n{sig}\n-----END {net_name} SIGNED MESSAGE-----')

    def __init__(self, network, generator):
        self._network = network
        self._network_name = network.network_name
        self._generator = generator

    @classmethod
    def parse_sections(class_, msg_in):
        # Convert to Unix line feeds from DOS style, iff we find them, but
        # restore to same at the end. The RFC implies we should be using
        # DOS \r\n in the message, but that does not always happen in today's
        # world of MacOS and Linux devs. A mix of types will not work here.
        dos_nl = ('\r\n' in msg_in)
        if dos_nl:
            msg_in = msg_in.replace('\r\n', '\n')

        try:
            # trim any junk in front
            _, body = msg_in.split('SIGNED MESSAGE-----\n', 1)
        except ValueError:
            raise EncodingError("expecting text SIGNED MESSSAGE somewhere")

        # - sometimes middle sep is BEGIN BITCOIN SIGNATURE, other times just BEGIN SIGNATURE
        # - choose the last instance, in case someone signs a signed message
        parts = re.split('\n-----BEGIN [A-Z ]*SIGNATURE-----\n', body)
        if len(parts) < 2:
            raise EncodingError("expected BEGIN SIGNATURE line", body)
        msg, hdr = ''.join(parts[:-1]), parts[-1]

        if dos_nl:
            msg = msg.replace('\n', '\r\n')

        return msg, hdr

    @classmethod
    def parse_signed_message(class_, msg_in):
        """
        Take an "armoured" message and split into the message body, signing address
        and the base64 signature. Should work on all altcoin networks, and should
        accept both Inputs.IO and Multibit formats but not Armory.

        Looks like RFC2550 <https://www.ietf.org/rfc/rfc2440.txt> was an "inspiration"
        for this, so in case of confusion it's a reference, but I've never found
        a real spec for this. Should be a BIP really.
        """

        msg, hdr = class_.parse_sections(msg_in)

        # after message, expect something like an email/http headers, so split into lines
        hdr = list(filter(None, [i.strip() for i in hdr.split('\n')]))

        if '-----END' not in hdr[-1]:
            raise EncodingError("expecting END on last line")

        sig = hdr[-2]
        addr = None
        for line in hdr:
            line = line.strip()
            if not line:
                continue

            if line.startswith('-----END'):
                break

            if ':' in line:
                label, value = [i.strip() for i in line.split(':', 1)]

                if label.lower() == 'address':
                    addr = line.split(':')[1].strip()
                    break

                continue

            addr = line
            break

        if not addr or addr == sig:
            raise EncodingError("Could not find address")

        return msg, addr, sig

    def signature_for_message_hash(self, secret_exponent, msg_hash, is_compressed):
        """
        Return a signature, encoded in Base64, of msg_hash.
        """
        r, s, recid = self._generator.sign_with_recid(secret_exponent, msg_hash)

        # See http://bitcoin.stackexchange.com/questions/14263 and key.cpp
        # for discussion of the proprietary format used for the signature

        first = 27 + recid + (4 if is_compressed else 0)
        sig = b2a_base64(int2byte(first) + to_bytes_32(r) + to_bytes_32(s)).strip()
        sig = sig.decode("utf8")
        return sig

    def sign_message(self, key, message, verbose=False):
        """
        Return a signature, encoded in Base64, which can be verified by anyone using the
        public key.
        """
        secret_exponent = key.secret_exponent()
        if not secret_exponent:
            raise ValueError("Private key is required to sign a message")

        addr = key.address()

        msg_hash = self.hash_for_signing(message)
        is_compressed = key.is_compressed()

        sig = self.signature_for_message_hash(secret_exponent, msg_hash, is_compressed)

        if not verbose or message is None:
            return sig

        return self.signature_template.format(
            msg=message, sig=sig, addr=addr,
            net_name=self._network_name.upper())

    def pair_for_message_hash(self, signature, msg_hash):
        """
        Take a signature, encoded in Base64, and return the pair it was signed by.
        May raise EncodingError (from _decode_signature)
        """

        # Decode base64 and a bitmask in first byte.
        is_compressed, recid, r, s = self._decode_signature(signature)

        # Calculate the specific public key used to sign this message.
        y_parity = recid & 1
        q = self._generator.possible_public_pairs_for_signature(msg_hash, (r, s), y_parity=y_parity)[0]
        if recid > 1:
            order = self._generator.order()
            q = self._generator.Point(q[0] + order, q[1])
        return q, is_compressed

    def pair_matches_key(self, pair, key, is_compressed):
        # Check signing public pair is the one expected for the signature. It must be an
        # exact match for this key's public pair... or else we are looking at a validly
        # signed message, but signed by some other key.
        #
        if hasattr(key, "public_pair"):
            # expect an exact match for public pair.
            return key.public_pair() == pair
        else:
            # Key() constructed from a hash of pubkey doesn't know the exact public pair, so
            # must compare hashed addresses instead.
            key_hash160 = key.hash160()
            pair_hash160 = public_pair_to_hash160_sec(pair, compressed=is_compressed)
            return key_hash160 == pair_hash160

    def verify_message(self, key_or_address, signature, message=None, msg_hash=None):
        """
        Take a signature, encoded in Base64, and verify it against a
        key object (which implies the public key),
        or a specific base58-encoded pubkey hash.
        """
        if isinstance(key_or_address, str):
            # they gave us a private key or a public key already loaded.
            key = self._network.parse.address(key_or_address)
        else:
            key = key_or_address

        try:
            msg_hash = self.hash_for_signing(message) if message is not None else msg_hash
            pair, is_compressed = self.pair_for_message_hash(signature, msg_hash)
        except EncodingError:
            return False
        return self.pair_matches_key(pair, key, is_compressed)

    def msg_magic_for_netcode(self):
        """
        We need the constant "strMessageMagic" in C++ source code, from file "main.cpp"

        It is not shown as part of the signed message, but it is prefixed to the message
        as part of calculating the hash of the message (for signature). It's also what
        prevents a message signature from ever being a valid signature for a transaction.

        Each altcoin finds and changes this string... But just simple substitution.
        """
        return '%s Signed Message:\n' % self._network_name

    def _decode_signature(self, signature):
        """
            Decode the internal fields of the base64-encoded signature.
        """

        sig = a2b_base64(signature)
        if len(sig) != 65:
            raise EncodingError("Wrong length, expected 65")

        # split into the parts.
        first = byte2int(sig)
        r = from_bytes_32(sig[1:33])
        s = from_bytes_32(sig[33:33+32])

        # first byte encodes a bits we need to know about the point used in signature
        if not (27 <= first < 35):
            raise EncodingError("First byte out of range")

        # NOTE: The first byte encodes the "recovery id", or "recid" which is a 3-bit values
        # which selects compressed/not-compressed and one of 4 possible public pairs.
        #
        first -= 27
        is_compressed = bool(first & 0x4)

        return is_compressed, (first & 0x3), r, s

    def hash_for_signing(self, msg):
        """
        Return a hash of msg, according to odd bitcoin method: double SHA256 over a bitcoin
        encoded stream of two strings: a fixed magic prefix and the actual message.
        """
        magic = self.msg_magic_for_netcode()

        fd = io.BytesIO()
        stream_satoshi_string(fd, magic.encode('utf8'))
        stream_satoshi_string(fd, msg.encode('utf8'))

        # return as a number, since it's an input to signing algos like that anyway
        return from_bytes_32(double_sha256(fd.getvalue()))




import io

from .encoding.hash import double_sha256
from .encoding.hexbytes import b2h, b2h_rev
from .merkle import merkle
from .satoshi.satoshi_struct import parse_struct, stream_struct


class BadMerkleRootError(Exception):
    pass


def difficulty_max_mask_for_bits(bits):
    prefix = bits >> 24
    mask = (bits & 0x7ffff) << (8 * (prefix - 3))
    return mask


class Block(object):
    """A Block is an element of the Bitcoin chain."""

    @classmethod
    def make_subclass(class_, tx):

        class Block(class_):
            Tx = tx

        return Block

    @classmethod
    def parse(class_, f, include_transactions=True, include_offsets=None, check_merkle_hash=True):
        """
        Parse the Block from the file-like object
        """
        block = class_.parse_as_header(f)
        if include_transactions:
            count = parse_struct("I", f)[0]
            txs = block._parse_transactions(f, count, include_offsets=include_offsets)
            block.set_txs(txs, check_merkle_hash=check_merkle_hash)
        return block

    @classmethod
    def parse_as_header(class_, f):
        """
        Parse the Block header from the file-like object
        """
        (version, previous_block_hash, merkle_root, timestamp,
            difficulty, nonce) = parse_struct("L##LLL", f)
        return class_(version, previous_block_hash, merkle_root, timestamp, difficulty, nonce)

    @classmethod
    def from_bin(class_, bytes):
        f = io.BytesIO(bytes)
        return class_.parse(f)

    def __init__(self, version, previous_block_hash, merkle_root, timestamp, difficulty, nonce):
        self.version = version
        self.previous_block_hash = previous_block_hash
        self.merkle_root = merkle_root
        self.timestamp = timestamp
        self.difficulty = difficulty
        self.nonce = nonce
        self.txs = []

    def set_nonce(self, nonce):
        self.nonce = nonce
        if hasattr(self, "__hash"):
            del self.__hash

    def _calculate_hash(self):
        s = io.BytesIO()
        self.stream_header(s)
        return double_sha256(s.getvalue())

    def hash(self):
        """Calculate the hash for the block header. Note that this has the bytes
        in the opposite order from how the header is usually displayed (so the
        long string of 00 bytes is at the end, not the beginning)."""
        if not hasattr(self, "__hash"):
            self.__hash = self._calculate_hash()
        return self.__hash

    @classmethod
    def _parse_transactions(class_, f, count, include_offsets=None):
        txs = []
        for i in range(count):
            if include_offsets:
                offset_in_block = f.tell()
            tx = class_.Tx.parse(f)
            txs.append(tx)
            if include_offsets:
                tx.offset_in_block = offset_in_block
        return txs

    def set_txs(self, txs, check_merkle_hash=True):
        self.txs = txs
        if not txs:
            return
        for tx in txs:
            tx.block = self
        if check_merkle_hash:
            self.check_merkle_hash()

    def as_blockheader(self):
        return Block(self.version, self.previous_block_hash, self.merkle_root,
                     self.timestamp, self.difficulty, self.nonce)

    def stream_header(self, f):
        """Stream the block header in the standard way to the file-like object f."""
        stream_struct("L##LLL", f, self.version, self.previous_block_hash,
                      self.merkle_root, self.timestamp, self.difficulty, self.nonce)

    def _stream_transactions(self, f):
        if self.txs:
            stream_struct("I", f, len(self.txs))
            for tx in self.txs:
                tx.stream(f)

    def stream(self, f):
        """Stream the block header in the standard way to the file-like object f.
        The Block subclass also includes the transactions."""
        self.stream_header(f)
        self._stream_transactions(f)

    def as_bin(self):
        """Return the block (or header) as binary."""
        f = io.BytesIO()
        self.stream(f)
        return f.getvalue()

    def as_hex(self):
        """Return the block (or header) as hex."""
        return b2h(self.as_bin())

    def id(self):
        """Returns the hash of the block displayed with the bytes in the order
        they are usually displayed in."""
        return b2h_rev(self.hash())

    def previous_block_id(self):
        """Returns the hash of the previous block, with the bytes in the order
        they are usually displayed in."""
        return b2h_rev(self.previous_block_hash)

    def check_merkle_hash(self):
        """Raise a BadMerkleRootError if the Merkle hash of the
        transactions does not match the Merkle hash included in the block."""
        calculated_hash = merkle([tx.hash() for tx in self.txs], double_sha256)
        if calculated_hash != self.merkle_root:
            raise BadMerkleRootError(
                "calculated %s but block contains %s" % (b2h(calculated_hash), b2h(self.merkle_root)))

    def __str__(self):
        c = '%s%s' % (self.__class__.__name__, '' if self.txs else 'Header')
        return "%s [%s] (previous %s)" % (c, self.id(), self.previous_block_id())

    def __repr__(self):
        return self.__str__()



import hashlib
import unittest

from pycoin.encoding.hexbytes import b2h
from pycoin.symbols.btc import network


# BRAIN DAMAGE
Tx = network.tx
TxIn = Tx.TxIn
TxOut = Tx.TxOut
Solver = Tx.Solver

ScriptError = network.validator.ScriptError

class SolverTest(unittest.TestCase):

    def do_test_solve(self, tx, tx_in_idx, **kwargs):
        solver = Solver(tx)
        constraints = solver.determine_constraints(tx_in_idx, p2sh_lookup=kwargs.get("p2sh_lookup"))
        solution_list, witness_list = solver.solve_for_constraints(constraints, **kwargs)
        solution_script = network.script.compile_push_data_list(solution_list)
        tx.txs_in[tx_in_idx].script = solution_script
        tx.txs_in[tx_in_idx].witness = witness_list
        if not kwargs.get("nocheck"):
            tx.check_solution(tx_in_idx)
        return solution_script, witness_list

    def make_test_tx(self, input_script):
        previous_hash = b'\1' * 32
        txs_in = [TxIn(previous_hash, 0)]
        txs_out = [TxOut(1000, network.contract.for_address(network.keys.private(1).address()))]
        version, lock_time = 1, 0
        tx = Tx(version, txs_in, txs_out, lock_time)
        unspents = [TxOut(1000, input_script)]
        tx.set_unspents(unspents)
        return tx

    def do_test_tx(self, incoming_script, **kwargs):
        keys = [network.keys.private(i) for i in range(1, 20)]
        tx = self.make_test_tx(incoming_script)
        tx_in_idx = 0
        kwargs["hash160_lookup"] = network.tx.solve.build_hash160_lookup((k.secret_exponent() for k in keys))
        kwargs["generator_for_signature_type_f"] = Solver.SolutionChecker.VM.generator_for_signature_type
        self.do_test_solve(tx, tx_in_idx, **kwargs)

    def test_p2pkh(self):
        key = network.keys.private(1)
        self.do_test_tx(network.contract.for_address(key.address()))

    def test_p2pk(self):
        key = network.keys.private(1)
        self.do_test_tx(network.contract.for_p2pk(key.sec(is_compressed=False)))
        self.do_test_tx(network.contract.for_p2pk(key.sec(is_compressed=True)))

    def test_nonstandard_p2pkh(self):
        key = network.keys.private(1)
        self.do_test_tx(network.script.compile("OP_SWAP") + network.contract.for_address(key.address()))

    def test_p2multisig(self):
        keys = [network.keys.private(i) for i in (1, 2, 3)]
        secs = [k.sec() for k in keys]
        self.do_test_tx(network.contract.for_multisig(2, secs))

    def test_p2sh(self):
        keys = [network.keys.private(i) for i in (1, 2, 3)]
        secs = [k.sec() for k in keys]
        underlying_script = network.contract.for_multisig(1, secs)
        script = network.contract.for_address(network.address.for_p2s(underlying_script))
        self.do_test_tx(script, p2sh_lookup=network.tx.solve.build_p2sh_lookup([underlying_script]))

        underlying_script = network.script.compile("OP_SWAP") + network.contract.for_address(keys[0].address())
        script = network.contract.for_address(network.address.for_p2s(underlying_script))
        self.do_test_tx(script, p2sh_lookup=network.tx.solve.build_p2sh_lookup([underlying_script]))

        underlying_script = network.contract.for_p2pk(keys[2].sec())
        script = network.contract.for_address(network.address.for_p2s(underlying_script))
        self.do_test_tx(script, p2sh_lookup=network.tx.solve.build_p2sh_lookup([underlying_script]))

    def test_p2pkh_wit(self):
        key = network.keys.private(1)
        script = network.script.compile("OP_0 [%s]" % b2h(key.hash160()))
        self.do_test_tx(script)

    def test_p2sh_wit(self):
        keys = [network.keys.private(i) for i in (1, 2, 3)]
        secs = [k.sec() for k in keys]
        underlying_script = network.contract.for_multisig(2, secs)
        script = network.script.compile("OP_0 [%s]" % b2h(hashlib.sha256(underlying_script).digest()))
        self.do_test_tx(script, p2sh_lookup=network.tx.solve.build_p2sh_lookup([underlying_script]))

    def test_p2multisig_wit(self):
        keys = [network.keys.private(i) for i in (1, 2, 3)]
        secs = [k.sec() for k in keys]
        underlying_script = network.contract.for_multisig(2, secs)
        p2sh_script = network.script.compile("OP_0 [%s]" % b2h(hashlib.sha256(underlying_script).digest()))
        script = network.contract.for_address(network.address.for_p2s(p2sh_script))
        self.do_test_tx(script, p2sh_lookup=network.tx.solve.build_p2sh_lookup([underlying_script, p2sh_script]))

    def test_if(self):
        script = network.script.compile("IF 1 ELSE 0 ENDIF")
        # self.do_test_tx(script)

    def test_p2multisig_incremental(self):
        keys = [network.keys.private(i) for i in (1, 2, 3)]
        secs = [k.sec() for k in keys]
        tx = self.make_test_tx(network.contract.for_multisig(3, secs))
        tx_in_idx = 0
        for k in keys:
            try:
                tx.check_solution(tx_in_idx)
                assert 0
            except ScriptError:
                pass
            kwargs = {"hash160_lookup": network.tx.solve.build_hash160_lookup([k.secret_exponent()])}
            kwargs["existing_script"] = [
                data for opcode, data, pc, new_pc in network.script.get_opcodes(
                    tx.txs_in[tx_in_idx].script) if data is not None]
            kwargs["nocheck"] = True
            kwargs["generator_for_signature_type_f"] = Solver.SolutionChecker.VM.generator_for_signature_type
            solution_list, witness_list = self.do_test_solve(tx, tx_in_idx, **kwargs)
        tx.check_solution(tx_in_idx)


if __name__ == "__main__":
    unittest.main()



import hashlib
import hmac

from . import intstream


if hasattr(1, "bit_length"):
    def bit_length(v):
        "the ``int.bit_length`` in `python 3 <https://docs.python.org/3/library/stdtypes.html#int.bit_length>`_"
        return v.bit_length()
else:
    def bit_length(self):
        "the ``int.bit_length`` in `python 3 <https://docs.python.org/3/library/stdtypes.html#int.bit_length>`_"
        # compared to "while n>0: bl +=1 ; n >>= 1", this is much faster in both python2 and pypy
        # code taken from the link above
        s = bin(self)  # binary representation:  bin(-37) --> '-0b100101'
        s = s.lstrip('-0b')  # remove leading zeros and minus sign
        return len(s)  # len('100101') --> 6


def deterministic_generate_k(generator_order, secret_exponent, val, hash_f=hashlib.sha256):
    """
    :param generator_order: result from `pycoin.ecdsa.Generator.Generator.order`,
        necessary to ensure the k value is within bound
    :param secret_exponent: an integer secret_exponent to generate the k value for
    :param val: the value to be signed, also used as an entropy source for the k value
    :returns: an integer k such that ``1 <= k < generator_order``, complying with
        <https://tools.ietf.org/html/rfc6979>
    """
    n = generator_order
    bln = bit_length(n)
    order_size = (bln + 7) // 8
    hash_size = hash_f().digest_size
    v = b'\x01' * hash_size
    k = b'\x00' * hash_size
    priv = intstream.to_bytes(secret_exponent, length=order_size)
    shift = 8 * hash_size - bln
    if shift > 0:
        val >>= shift
    if val >= n:
        val -= n
    h1 = intstream.to_bytes(val, length=order_size)
    k = hmac.new(k, v + b'\x00' + priv + h1, hash_f).digest()
    v = hmac.new(k, v, hash_f).digest()
    k = hmac.new(k, v + b'\x01' + priv + h1, hash_f).digest()
    v = hmac.new(k, v, hash_f).digest()

    while 1:
        t = bytearray()

        while len(t) < order_size:
            v = hmac.new(k, v, hash_f).digest()
            t.extend(v)

        k1 = intstream.from_bytes(bytes(t))

        k1 >>= (len(t)*8 - bln)
        if k1 >= 1 and k1 < n:
            return k1

        k = hmac.new(k, v + b'\x00', hash_f).digest()
        v = hmac.new(k, v, hash_f).digest()


"""
Some portions adapted from https://github.com/warner/python-ecdsa/ Copyright (c) 2010 Brian Warner
who granted its use under this license:

Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.


Portions written in 2005 by Peter Pearson and placed in the public domain.
"""



from pycoin.networks.bitcoinish import create_bitcoinish_network

from pycoin.coins.bcash.Tx import Tx as BcashTx


network = create_bitcoinish_network(
    symbol="XCH", network_name="Bcash", subnet_name="testnet3", tx=BcashTx,
    wif_prefix_hex="ef", sec_prefix="XCHSEC:", address_prefix_hex="6f", pay_to_script_prefix_hex="c4",
    bip32_prv_prefix_hex="04358394", bip32_pub_prefix_hex="043587CF", bech32_hrp="tb",
    magic_header_hex="0B110907", default_port=18333,
    dns_bootstrap=[
        "seed.bitcoinabc.org", "seed-abc.bitcoinforks.org",
        "btccash-seeder.bitcoinunlimited.info", "seed.bitprim.org",
        "seed.deadalnix.me", "seeder.criptolayer.net"
    ])



from pycoin.networks.bitcoinish import create_bitcoinish_network

network = create_bitcoinish_network(
    symbol="JBS", network_name="Jumbucks", subnet_name="mainnet",
    wif_prefix_hex="ab", address_prefix_hex="2b",
    bip32_prv_prefix_hex="037a6460", bip32_pub_prefix_hex="037a689a")
